package ch.polgrabia.demos.client;

import ch.polgrabia.demos.utils.TopicUtil;
import io.vertx.core.AbstractVerticle;
import io.vertx.core.AsyncResult;
import io.vertx.core.http.HttpClient;
import io.vertx.core.http.HttpClientOptions;
import io.vertx.core.http.WebSocket;
import io.vertx.core.http.impl.WebSocketClientImpl;
import io.vertx.core.impl.CloseFuture;
import io.vertx.core.impl.VertxInternal;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.CompletableFuture;

public class WebsocketClientParticle extends AbstractVerticle {
  private static final Logger logger = LoggerFactory.getLogger(WebsocketClientParticle.class);
  private final String hostname;
  private final int port;
  // path maps to channel
  private final String path;
  private final String channel;
  private final ClientPeriodicStdinScannerParticle clientPeriodicStdinScannerParticle = new ClientPeriodicStdinScannerParticle(
    System.in
  );
  private final WebApiParticle webApiParticle = new WebApiParticle();
  private HttpClient client;
  private boolean isShuttingDownActivated = false;
  private final Object runningMonitor = new Object();
  private WebSocket webSocket;
  private WebSocketSenderParticle webSocketSenderParticle;

  public WebsocketClientParticle(String hostname, int port, String path) {
    this.hostname = hostname;
    this.port = port;
    this.path = path;
    this.channel = TopicUtil.convertTopicName(path);
  }

  @Override
  public void start() {
    this.client = getVertx()
      .createHttpClient();
    logger.info("Starting websocket client command server");
    new WebSocketClientImpl((VertxInternal) getVertx(), new HttpClientOptions(), new CloseFuture())
      .webSocket(port, hostname, path, this::handleMessage);
    getVertx().deployVerticle(clientPeriodicStdinScannerParticle);
    getVertx().deployVerticle(webApiParticle);
  }

  private void handleMessage(AsyncResult<WebSocket> webSocketAsyncResult) {
    logger.info("Handling websocket");
    if (!webSocketAsyncResult.succeeded()) {
      logger.warn("Websocket handler connection failed", webSocketAsyncResult.cause());
      try {
        stop();
      } catch (Exception e) {
        logger.error("Failed to stop particle", e);
      }
      return;
    }
    this.webSocket = webSocketAsyncResult.result();
    webSocketSenderParticle = new WebSocketSenderParticle(webSocket);
    getVertx().deployVerticle(webSocketSenderParticle);
    vertx.setPeriodic(100, clientPeriodicStdinScannerParticle);

    webSocket
      .binaryMessageHandler(msg -> logger.info("[{}] Got message: {}", channel, msg))
      .closeHandler(unused -> {
        try {
          stop();
        } catch (Exception e) {
          logger.error("Got an error while stopping", e);
        }
      });

  }

  @Override
  public void stop() {
    isShuttingDownActivated = true;
    getVertx().undeploy(webApiParticle.deploymentID());
    getVertx().undeploy(webSocketSenderParticle.deploymentID());
    getVertx().undeploy(clientPeriodicStdinScannerParticle.deploymentID());
    client.close();
    synchronized (runningMonitor) {
      runningMonitor.notifyAll();
    }
  }

  public CompletableFuture<Boolean> waitingUntilFinished() {
    return CompletableFuture.supplyAsync(() -> {
      while (!isShuttingDownActivated) {
        synchronized (runningMonitor) {
          try {
            runningMonitor.wait();
            return true;
          } catch (InterruptedException e) {
            logger.warn("Got interrupted", e);
          }
        }
      }
      return false;
    });
  }

}
